from copy import deepcopy
from typing import Union, Set, Deque, Dict, List
from ib110hw.automaton.nfa import NFA
from ib110hw.automaton.dfa import DFA, DFATransitions
from collections import deque


def determinize(automaton: NFA) -> DFA:
    """
    Returns equivalent DFA version of the provided NFA using the powerset construction.
    Input automaton is not altered.

    Args:
        automaton (NFA): NFA automaton to be determinized

    Returns:
        DFA: Determinized automaton
    """
    automaton = remove_empty_transitions(automaton)

    states: Deque[Set[str]] = deque()
    states.append({automaton.initial_state})

    det_states: Set[str] = set()
    det_final_states: Set[str] = set()
    det_transitions = {automaton.initial_state: {}}

    while states:
        state = states.popleft()
        str_state = "".join(sorted(state))

        if str_state in det_states:
            continue

        det_states.add(str_state)
        det_transitions[str_state] = {}

        if automaton.final_states.intersection(state):
            det_final_states.add(str_state)

        for key in automaton.alphabet:
            new_state = set()
            for s in state:
                new_state = new_state.union(automaton.get_transition(s, key))

            if not new_state:
                new_state = {"sink"}

            det_transitions[str_state][key] = "".join(sorted(new_state))
            states.append(new_state)

    return DFA(
        states=det_states,
        alphabet={*automaton.alphabet},
        final_states={*det_final_states},
        initial_state=automaton.initial_state,
        transitions=det_transitions,
    )


def remove_empty_transitions(automaton: NFA) -> NFA:
    """
    Removes all epsilon (ε) transitions.
    Input automaton is not altered.

    Args:
        automaton (NFA): Automaton

    Returns:
        NFA: Equivalent automaton without ε transitions.
    """

    result: NFA = NFA(
        {*automaton.states},
        automaton.alphabet.difference({""}),
        automaton.initial_state,
        {*automaton.final_states},
        {},
    )

    if "" not in automaton.alphabet:
        result.transitions = deepcopy(automaton.transitions)
        return result

    next1: Dict[str, Set[str]] = {}

    # next1
    for state in automaton.states:
        next1[state] = {state}.union(automaton.get_transition(state, ""))

        for next_state in automaton.get_transition(state, ""):
            next1[state] = next1[state].union(next1.get(next_state, set()))

    # next2
    for state in next1:
        for symbol in result.alphabet:
            for next_state in next1[state]:
                for s in automaton.get_transition(next_state, symbol):
                    result.add_transition(state, s, symbol)

    # next3
    for state in result.transitions:
        for symbol in result.alphabet:
            next2_transition = {*result.get_transition(state, symbol)}

            for next_state in next2_transition:
                for s in next1.get(next_state, set()):
                    result.add_transition(state, s, symbol)

    return result


def remove_unreachable_states(automaton: Union[DFA, NFA]) -> Union[NFA, DFA]:
    """
    Removes unreachable states from the provided automaton.
    Input automaton is not altered.

    Args:
        automaton (Union[DFA, NFA]): Automaton

    Returns:
        Equal automaton without unreachable states.
    """
    result = deepcopy(automaton)
    reachable = {automaton.initial_state}
    queue = deque()

    queue.append(automaton.initial_state)

    while queue:
        state = queue.popleft()

        if isinstance(automaton, DFA):
            next_state = set()
            for symbol in automaton.alphabet:
                next_state.update({automaton.get_transition(state, symbol)})
        else:
            next_state = {
                automaton.get_transition(state, symbol) for symbol in automaton.alphabet
            }

        queue.extend(next_state - reachable)
        reachable.update(next_state)

    for state in list(automaton.states - reachable):
        result.remove_state(state)

    return result


def minimize(automaton: Union[DFA, NFA]) -> DFA:
    """
    Returns a minimized version of the provided automaton.
    Input automaton is not altered.

    Args:
        automaton (Union[DFA, NFA]): Automaton to be minimized

    Returns:
        DFA: Minimized version of the provided automaton.
    """

    def get_groups(_transitions: DFATransitions) -> List[Set[str]]:
        """
        Divides states into equivalence groups.
        """
        new_groups = {}

        for g_index, _group in enumerate(groups):
            for _state in sorted(_group):
                # group key is prefixed with index
                # to distinguish the same transitions from different groups
                group_key = f"{g_index}_"

                if _state not in _transitions.keys():
                    continue

                for _symbol in automaton.alphabet:
                    transition = _transitions.get(_state, {}).get(_symbol, None)
                    if transition:
                        group_key += transition

                if group_key not in new_groups.keys():
                    new_groups[group_key] = set()

                new_groups[group_key].add(_state)

        return [new_groups[key] for key in sorted(new_groups.keys())]

    if isinstance(automaton, NFA):
        automaton = determinize(automaton)

    automaton: DFA = remove_unreachable_states(automaton)
    minimized_transitions: DFATransitions = {}

    result: DFA = DFA(
        {*automaton.states},
        {*automaton.alphabet},
        automaton.initial_state,
        set(),
        minimized_transitions,
    )

    groups: List[Set[str]] = [
        {*automaton.final_states},
        automaton.states.difference(automaton.final_states),
    ]

    while True:
        # create transitions with group indexes instead of states
        # break when nothing changes
        marked_transitions = {}

        for state in automaton.transitions:
            for symbol in sorted(automaton.alphabet):
                for index, group in enumerate(groups):
                    if automaton.get_transition(state, symbol) not in group:
                        continue

                    if state not in marked_transitions:
                        marked_transitions[state] = {}

                    marked_transitions[state][symbol] = f"{index}"

        prev_len = len(groups)
        groups = get_groups(marked_transitions)

        if prev_len == len(groups):
            break

    result.states = {f"{i}" for i in range(len(groups))}

    for index in range(len(groups)):
        if groups[index].intersection(automaton.final_states):
            result.final_states.add(f"{index}")

        if automaton.initial_state in groups[index]:
            result.initial_state = f"{index}"

        minimized_transitions[f"{index}"] = marked_transitions[groups[index].pop()]

    return result


def canonize(automaton: DFA) -> DFA:
    """
    Transforms provided automaton to its canonical form.
    Input automaton is not altered.

    Args:
        automaton (DFA): Automaton to be transformed.

    Returns:
        Canonical form of the provided automaton.
    """

    def append_next():
        next_state = next((s for s in automaton.transitions if s not in renamed), None)
        if next_state:
            states.append(next_state)

    renamed = {}
    states = deque()
    states.append(automaton.initial_state)

    # used for renaming the states
    # name is assigned based on the order the state was found
    # from the initial state
    rank = 0

    while states:
        current_state = states.popleft()

        if renamed.get(current_state):
            if states:
                continue
            if len(renamed) < len(automaton.states):
                append_next()
                continue

        renamed[current_state] = f"{rank}"

        for symbol in sorted(automaton.alphabet):
            next_state = automaton.get_transition(current_state, symbol)

            if renamed.get(next_state):
                continue

            states.append(next_state)

        rank += 1
        if not states:
            append_next()

    renamed_transitions = {
        renamed[state]: deepcopy(automaton.transitions[state])
        for state in automaton.states
    }

    for state, value in renamed_transitions.items():
        for symbol in value:
            prev = renamed_transitions[state][symbol]
            renamed_transitions[state][symbol] = renamed[prev]

    return DFA(
        {renamed[state] for state in automaton.states},
        {*automaton.alphabet},
        renamed[automaton.initial_state],
        {renamed[state] for state in automaton.final_states},
        renamed_transitions,
    )


def compare_automata(a1: Union[NFA, DFA], a2: Union[NFA, DFA]) -> bool:
    """
    Compares the two automata by transforming them into their canonical form.

    Args:
        a1 (Union[NFA, DFA]): Automaton to be compared.
        a2 (Union[NFA, DFA]): Automaton to be compared.

    Returns:
        bool: True if the provided automata are equivalent. False otheriwse.
    """
    if isinstance(a1, NFA):
        a1 = determinize(a1)

    if isinstance(a2, NFA):
        a2 = determinize(a2)

    a1 = canonize(minimize(a1))
    a2 = canonize(minimize(a2))

    return (
        a1.states == a2.states
        and a1.final_states == a2.final_states
        and a1.alphabet == a2.alphabet
        and a1.transitions == a2.transitions
    )


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ib110hw_testing', 'ib110hw_testing.testing', 'ib110hw_testing.transformation']

package_data = \
{'': ['*']}

install_requires = \
['exrex', 'hypothesis', 'ib110hw']

setup_kwargs = {
    'name': 'ib110hw-testing',
    'version': '0.1.11',
    'description': 'Package used for testing homework assignments in the IB110 course at FI MUNI.',
    'long_description': 'This library was created for the course **IB110 - Introduction to Informatics** at [MUNI FI](https://www.fi.muni.cz/). It builds on top of the [ib110hw](https://pypi.org/project/ib110hw/) and the [hypothesis](https://pypi.org/project/hypothesis/) libraries.\n\n# Setup\n\nPython version required for this library is **>=3.6**. It can be installed using `pip` like so:\n```pip install ib110hw_testing```\n\nUsing venv to install the library is recommended as it has some dependencies (`ib110hw`, `exrex`, `hypothesis`).\n\n# Modules\n\n## Testing\n\nThe module `testing` contains some predefined strategies which can be used with the `hypothesis` library.\n\n### Example Use-case\n\nBelow code shows a simple example of strategies used with tests.\n\n```python\nfrom hypothesis import given\nfrom ib110hw.automaton.dfa import DFA\nfrom ib110hw_testing.testing.strategies import dfas, strings_from_regex\nfrom typing import Set\n\n@given(\n    dfas(alphabet={"a", "b", "c"}),\n    strings_from_regex(regex="[abc]*")\n)\ndef test(dfa: DFA, strings: Set[str])\n    ...\n```\n\n## Transformation\n\nThe module `transformation` contains implementations of some algorithms which can be used to transform automata.\n\n### Example Use-case\n\nBelow code shows a simple example of automata comparison.\n\n```python\nfrom ib110hw.automaton.dfa import DFA\nfrom ib110hw.automaton.nfa import NFA\nfrom ib110hw_testing.transformation.transformation import compare_automata\n\nnfa: NFA = NFA(...) # an arbitrary NFA\ndfa: DFA = DFA(...) # an arbitrary DFA\n\ncompare_automata(nfa, dfa) # returns True/False\n```\n',
    'author': 'Martin Pilát',
    'author_email': '8pilatmartin8@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.8,<4.0.0',
}


setup(**setup_kwargs)



from exrex import generate
from hypothesis.strategies import integers, composite, DrawFn, lists, sampled_from
from ib110hw.automaton.dfa import DFA
from ib110hw.automaton.nfa import NFA
from itertools import islice
from typing import Set


@composite
def states(draw: DrawFn, min_states: int, max_states: int):
    """
    Composite strategy to generate automaton states with name formatted as 's' + index.

    Args:
        min_states (int): The minimum amount of states.
        max_states (int): The maximum amount of states.
    """
    return [
        f"s{i}"
        for i in range(draw(integers(min_value=min_states, max_value=max_states)))
    ]


@composite
def dfas(
    draw: DrawFn,
    alphabet: Set[str],
    min_states: int = 5,
    max_states: int = 10,
    min_fin_states: int = 1,
    max_fin_states: int = 10,
):
    """
    Composite strategy to generate random DFA based on the provided params.

    Args:
        alphabet (Set[str]): Alphabet of the automaton.
        min_states (int, optional): The minimum amount of states. Defaults to 5.
        max_states (int, optional): The maximum amount of states. Defaults to 10.
        min_fin_states (int, optional): The minimum amount of final states. Defaults to 1.
        max_fin_states (int, optional): The maximum amount of final states. Defaults to 10.
    """

    def add_next_states(_state: str) -> None:
        next_states = draw(
            lists(
                sampled_from(automaton_states),
                min_size=len(alphabet),
                max_size=len(alphabet),
            )
        )

        for next_s, symbol in zip(next_states, alphabet):
            automaton.add_transition(_state, next_s, symbol)

    automaton_states = draw(states(min_states, max_states))
    min_fin_size = min(len(automaton_states), min_fin_states)
    max_fin_size = min(len(automaton_states), max_fin_states)
    automaton = DFA(
        set(automaton_states),
        alphabet,
        draw(sampled_from(automaton_states)),
        set(
            draw(
                lists(
                    sampled_from(automaton_states),
                    min_size=min_fin_size,
                    max_size=max_fin_size,
                )
            )
        ),
        {},
    )

    add_next_states(automaton.initial_state)

    for state in automaton.states.difference([automaton.initial_state]):
        add_next_states(state)

    return automaton


@composite
def nfas(
    draw: DrawFn,
    alphabet: Set[str],
    min_deg: int = 1,
    max_deg: int = None,
    min_states: int = 5,
    max_states: int = 10,
    min_fin_states: int = 1,
    max_fin_states: int = 10,
):
    """
    Composite strategy to generate random DFA based on the provided params.

    Args:
        alphabet (Set[str]): Alphabet of the automaton.
        min_deg (int): The minimum amount of transitions from this state. Defaults to 1.
        max_deg (int): The maximum amount of transitions from this state. Defaults to the amount of states.
        min_states (int, optional): The minimum amount of states. Defaults to 5.
        max_states (int, optional): The maximum amount of states. Defaults to 10.
        min_fin_states (int, optional): The minimum amount of final states. Defaults to 1.
        max_fin_states (int, optional): The maximum amount of final states. Defaults to 10.
    """

    def add_next_states(_min_deg: int, _max_deg: int, _state: str) -> None:
        s_deg = integers(_min_deg, _max_deg)
        next_states = draw(
            lists(
                sampled_from(states), min_size=min(s_deg, len(states)), max_size=max_deg
            )
        )

        for next_state in next_states:
            symbol = draw(sampled_from(alphabet))
            automaton.add_transition(_state, {next_state}, symbol)

    automaton_states = draw(states(min_states, max_states))
    min_fin_size = min(len(automaton_states), min_fin_states)
    max_fin_size = min(len(automaton_states), max_fin_states)
    automaton = NFA(
        set(states),
        alphabet,
        draw(sampled_from(states)),
        set(
            draw(
                lists(
                    sampled_from(automaton_states),
                    min_size=min_fin_size,
                    max_size=max_fin_size,
                )
            )
        ),
        {},
    )

    if not (max_deg) or max_deg > len(automaton_states):
        max_deg = len(automaton_states)

    add_next_states(1, max_deg, automaton.initial_state)

    for state in automaton.states.difference([automaton.initial_state]):
        add_next_states(min_deg, max_deg, state)

    return automaton


@composite
def strings_from_regex(
    draw: DrawFn,
    regex: str,
    min_amount: int = 5,
    max_amount: int = 10,
    max_str_len: int = 5,
):
    """
    Composite strategy to generate strings based on the provided regular expression.

    Args:
        regex (str): Regular expression used to generate strings.
        min_amount (int, optional): Min amount of strings to be generated. Defaults to 5.
        max_amount (int, optional): Max amount of strings to be generated. Defaults to 10.
        max_str_len (int, optional): Max length of individual generated strings. Defaults to 5.
    """
    amount = draw(integers(min_value=min_amount, max_value=max_amount))
    generator = islice(generate(regex, limit=max_str_len), amount)

    return set(generator)


if __name__ == "__main__":
    pass
